"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g = Object.create((typeof Iterator === "function" ? Iterator : Object).prototype);
    return g.next = verb(0), g["throw"] = verb(1), g["return"] = verb(2), typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (g && (g = 0, op[0] && (_ = 0)), _) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.TagInput = void 0;
var Mustache = require("mustache");
// tslint:disable-next-line: no-empty
function noop() {
    var args = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        args[_i] = arguments[_i];
    }
}
function matches(el, selector) {
    return (el.matches || el.webkitMatchesSelector).call(el, selector);
}
var Key;
(function (Key) {
    Key["ENTER"] = "Enter";
    Key["TAB"] = "Tab";
    Key["BACKSPACE"] = "Backspace";
    Key["HOME"] = "Home";
    Key["END"] = "End";
    Key["LEFT_ARROW"] = "ArrowLeft";
    Key["UP_ARROW"] = "ArrowUp";
    Key["RIGHT_ARROW"] = "ArrowRight";
    Key["DOWN_ARROW"] = "ArrowDown";
})(Key || (Key = {}));
var KeyNavDirection;
(function (KeyNavDirection) {
    KeyNavDirection[KeyNavDirection["UP"] = 0] = "UP";
    KeyNavDirection[KeyNavDirection["DOWN"] = 1] = "DOWN";
})(KeyNavDirection || (KeyNavDirection = {}));
var EmptyString = '';
var controlKeys = [
    Key.ENTER,
    Key.TAB,
    Key.BACKSPACE,
    Key.HOME,
    Key.END,
    Key.LEFT_ARROW,
    Key.UP_ARROW,
    Key.RIGHT_ARROW,
    Key.DOWN_ARROW
];
// Concatenate the key ranges for numbers and letters
var alphaKeys = 'abcdefghijklmnopqrstuvwxyz';
var alphaKeysLower = alphaKeys.split('');
var alphaKeysUpper = alphaKeys.toUpperCase().split('');
var numericKeys = '0123456789'.split('');
// The only symbol allowed by default is minus (-), which can be used to separate words in a tag
var symbolKeys = ['-'];
var standardValidTagCharacters = alphaKeysLower
    .concat(alphaKeysUpper)
    .concat(numericKeys)
    .concat(symbolKeys)
    .concat(controlKeys);
var standardHtmlTemplate = "<div class=\"{{globalCssClassPrefix}}{{#containerClasses}} {{containerClasses}}{{/containerClasses}}\">\n    {{#tags}}{{> item}}{{/tags}}\n    <input class=\"{{globalCssClassPrefix}}-data\" type=\"hidden\" name=\"{{hiddenInput.name}}\" id=\"{{hiddenInput.id}}\" value=\"{{hiddenInput.value}}\">\n    <div class=\"{{globalCssClassPrefix}}-input-container\">\n        <input class=\"{{globalCssClassPrefix}}-input\" type=\"text\"{{#placeholder}} placeholder=\"{{placeholder}}\"{{/placeholder}}>\n        <div class=\"{{globalCssClassPrefix}}-suggestions {{globalCssClassPrefix}}-suggestions-hidden\"></div>\n    </div>\n</div>";
var standardItemTemplate = '<div class="{{globalCssClassPrefix}}-tag" data-id="{{id}}" data-label="{{label}}">{{label}} <i class="{{globalCssClassPrefix}}-removetag fa fa-times"></i></div>';
var standardSuggestionTemplate = '<div class="{{globalCssClassPrefix}}-suggestion" data-id="{{id}}" data-label="{{label}}">{{label}}</div>';
var TagInput = /** @class */ (function () {
    function TagInput(options) {
        var _this = this;
        this.getId = options.getId;
        this.getLabel = options.getLabel;
        this.newItemFactory = options.newItemFactory;
        this.data = options.data.map(function (item) { return ({ id: _this.getId(item), label: _this.getLabel(item) }); });
        var originalInput = options.input;
        this.enableSuggestions = options.enableSuggestions === false ? false : true;
        this.minCharsBeforeShowingSuggestions = options.minCharsBeforeShowingSuggestions || 2;
        this.maxNumberOfSuggestions = options.maxNumberOfSuggestions || 100;
        this.allowNewTags = options.allowNewTags === false ? false : true;
        if (this.allowNewTags && !this.newItemFactory) {
            throw Error('A newItemFactory function must be specified if allowNewTags is true');
        }
        this.tagDataSeparator = options.tagDataSeparator || '|';
        this.validTagCharacters = options.validTagCharacters || standardValidTagCharacters;
        this.allowUpperCase = options.allowUpperCase || false;
        this.onTagAdded = options.onTagAdded || noop;
        this.onTagRemoved = options.onTagRemoved || noop;
        this.onTagsChanged = options.onTagsChanged || noop;
        this.globalCssClassPrefix = options.globalCssClassPrefix || 'mab-bootstrap-taginput';
        this.placeholderText = originalInput.getAttribute('placeholder');
        this.htmlTemplate = (options.htmlTemplate || standardHtmlTemplate).trim();
        this.itemTemplate = (options.itemTemplate || standardItemTemplate).trim();
        this.suggestionTemplate = (options.suggestionTemplate || standardSuggestionTemplate).trim();
        Mustache.parse(this.htmlTemplate);
        Mustache.parse(this.itemTemplate);
        Mustache.parse(this.suggestionTemplate);
        this.warningClass = "".concat(this.globalCssClassPrefix, "-tag-warning");
        this.narrowedInputClass = "".concat(this.globalCssClassPrefix, "-input-narrowed");
        this.multiCaseInputClass = "".concat(this.globalCssClassPrefix, "-input-multicase");
        this.suggestionDropdownClass = "".concat(this.globalCssClassPrefix, "-suggestions");
        this.suggestionDropdownHiddenClass = "".concat(this.globalCssClassPrefix, "-suggestions-hidden");
        this.suggestionClass = "".concat(this.globalCssClassPrefix, "-suggestion");
        this.selectedSuggestionClass = "".concat(this.suggestionClass, "-selected");
        var selectedIds = originalInput.value.split(this.tagDataSeparator);
        this.currentSelection = this.data.filter(function (item) { return selectedIds.includes(item.id); });
        var templateData = {
            containerClasses: originalInput.classList.value,
            globalCssClassPrefix: this.globalCssClassPrefix,
            tags: this.currentSelection,
            hiddenInput: {
                name: originalInput.name,
                id: originalInput.id,
                value: originalInput.value,
            },
            placeholder: this.placeholderText
        };
        var partials = {
            item: this.itemTemplate
        };
        var html = Mustache.render(this.htmlTemplate, templateData, partials);
        var template = document.createElement('template');
        template.insertAdjacentHTML('afterbegin', html);
        this.tagInputContainer = template.firstChild.cloneNode(true);
        this.tagInputHiddenInput = this.tagInputContainer.querySelector('[type=hidden]');
        this.tagInputTextInput = this.tagInputContainer.querySelector('[type=text]');
        this.tagInputSuggestionDropdown = this.tagInputContainer.querySelector(".".concat(this.suggestionDropdownClass));
        template.remove();
        originalInput.insertAdjacentElement('afterend', this.tagInputContainer);
        originalInput.remove();
        if (this.allowUpperCase) {
            this.tagInputTextInput.classList.add(this.multiCaseInputClass);
        }
        this.tagInputContainer.addEventListener('click', function (e) {
            var element = e.target;
            if (element.classList.contains("".concat(_this.globalCssClassPrefix, "-removetag"))) {
                e.stopPropagation();
                _this.removeTag(element.parentElement);
            }
        }, false);
        if (this.enableSuggestions) {
            this.tagInputSuggestionDropdown.addEventListener('mousedown', function (e) { return __awaiter(_this, void 0, void 0, function () {
                return __generator(this, function (_a) {
                    e.preventDefault();
                    return [2 /*return*/];
                });
            }); });
            this.tagInputSuggestionDropdown.addEventListener('click', function (e) { return __awaiter(_this, void 0, void 0, function () {
                var element, label;
                return __generator(this, function (_a) {
                    switch (_a.label) {
                        case 0:
                            element = e.target;
                            label = element.getAttribute('data-label');
                            return [4 /*yield*/, this.addTag(label)];
                        case 1:
                            _a.sent();
                            return [2 /*return*/];
                    }
                });
            }); });
            this.tagInputTextInput.addEventListener('keyup', function (e) {
                // If the up or down arrows are hit, select the previous/next item in the suggestion list
                if (e.key === Key.UP_ARROW || e.key === Key.DOWN_ARROW) {
                    var direction = e.key === Key.DOWN_ARROW ? KeyNavDirection.DOWN : KeyNavDirection.UP;
                    var selectedItem = _this.tagInputSuggestionDropdown.querySelector(".".concat(_this.selectedSuggestionClass));
                    if (selectedItem) {
                        // TODO: Filter these based on class
                        // https://plainjs.com/javascript/traversing/get-siblings-of-an-element-40/
                        // An item is already selected, so select the previous/next sibling
                        var sibling = direction === KeyNavDirection.DOWN
                            ? selectedItem.nextElementSibling
                            : selectedItem.previousElementSibling;
                        // If sibling is null, we've reached the end (or beginning) of the list,
                        // so skip back around to the first (or last) item
                        if (!sibling) {
                            var allSiblings = _this.tagInputSuggestionDropdown.querySelectorAll(".".concat(_this.suggestionClass));
                            if (direction === KeyNavDirection.DOWN) {
                                sibling = allSiblings[0];
                            }
                            else {
                                sibling = allSiblings[allSiblings.length - 1];
                            }
                        }
                        selectedItem.classList.remove(_this.selectedSuggestionClass);
                        sibling.classList.add(_this.selectedSuggestionClass);
                        _this.tagInputTextInput.value = sibling.getAttribute('data-label');
                    }
                    else if (direction === KeyNavDirection.DOWN) {
                        var firstElement = _this.tagInputSuggestionDropdown.firstChild;
                        firstElement.classList.add(_this.selectedSuggestionClass);
                        _this.tagInputTextInput.value = firstElement.getAttribute('data-label');
                    }
                }
                else if (_this.validTagCharacters.includes(e.key)) {
                    var inputValue = _this.getTextInputValue(_this.allowUpperCase);
                    if (inputValue.length < _this.minCharsBeforeShowingSuggestions) {
                        _this.hideSuggestions();
                        return;
                    }
                    var searchValue = inputValue.toUpperCase();
                    var suggestions = _this.searchTags(_this.data, searchValue)
                        .map(function (t) {
                        var suggestionData = {
                            globalCssClassPrefix: _this.globalCssClassPrefix,
                            id: t.id,
                            label: t.label
                        };
                        return Mustache.render(_this.suggestionTemplate, suggestionData);
                    });
                    // TODO: show message (hit enter to add new tag) if no suggestions and new tags allowed?
                    if (!suggestions.length) {
                        _this.hideSuggestions();
                        return;
                    }
                    _this.tagInputSuggestionDropdown.innerHTML = suggestions
                        .slice(0, _this.maxNumberOfSuggestions)
                        .join('');
                    _this.showSuggestions();
                }
            });
        }
        this.tagInputTextInput.addEventListener('keydown', function (e) { return __awaiter(_this, void 0, void 0, function () {
            var inputValue, tagElements, lastTagElement;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        if (!this.validTagCharacters.includes(e.key)) {
                            e.preventDefault();
                        }
                        inputValue = this.getTextInputValue(this.allowUpperCase);
                        if (!(e.key === Key.ENTER && inputValue !== EmptyString)) return [3 /*break*/, 2];
                        // Stop the form being submitted and prevent event bubbling
                        e.preventDefault();
                        e.stopPropagation();
                        return [4 /*yield*/, this.addTag(inputValue)];
                    case 1:
                        _a.sent();
                        _a.label = 2;
                    case 2:
                        // If backspace is hit and there's nothing in the input (if the input *isn't* empty,
                        // we don't want to prevent the default action, which is deleting a character)
                        if (e.key === Key.BACKSPACE && inputValue === EmptyString) {
                            tagElements = this.tagInputContainer.querySelectorAll(".".concat(this.globalCssClassPrefix, "-tag"));
                            if (tagElements.length) {
                                lastTagElement = tagElements[tagElements.length - 1];
                                this.removeTag(lastTagElement);
                            }
                        }
                        return [2 /*return*/];
                }
            });
        }); });
        this.tagInputTextInput.addEventListener('paste', function (e) { return __awaiter(_this, void 0, void 0, function () {
            var pastedText, newTags, _i, newTags_1, tag;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        pastedText = e.clipboardData.getData('text');
                        newTags = pastedText
                            .replace(/\s{0,}[,| ]\s{0,}/gi, '|')
                            .split('|');
                        _i = 0, newTags_1 = newTags;
                        _a.label = 1;
                    case 1:
                        if (!(_i < newTags_1.length)) return [3 /*break*/, 4];
                        tag = newTags_1[_i];
                        return [4 /*yield*/, this.addTag(tag)];
                    case 2:
                        _a.sent();
                        _a.label = 3;
                    case 3:
                        _i++;
                        return [3 /*break*/, 1];
                    case 4:
                        e.preventDefault();
                        return [2 /*return*/];
                }
            });
        }); });
        function isTextInput(target) {
            var element = target;
            return matches(element, 'input[type=text]');
        }
        this.tagInputTextInput.addEventListener('focus', function (e) {
            if (isTextInput(e.target)) {
                _this.showTagInput();
            }
        });
        this.tagInputTextInput.addEventListener('blur', function (e) {
            if (isTextInput(e.target)) {
                _this.clearTagInput();
                _this.hideSuggestions();
                if (_this.currentSelection.length) {
                    _this.hideTagInput();
                }
            }
        });
        // Focus the text input when the control container is clicked, which triggers
        // the show/hide behaviours defined in the handlers above
        this.tagInputContainer.addEventListener('click', function (e) {
            _this.focus();
        });
        // If the control already has some tags in it, hide the text input on load
        if (this.getHiddenInputValue() !== EmptyString) {
            this.hideTagInput();
        }
    }
    TagInput.prototype.getValue = function () {
        return this.getHiddenInputValue();
    };
    TagInput.prototype.focus = function () {
        this.tagInputTextInput.focus();
    };
    TagInput.prototype.updateData = function (data) {
        var _a;
        var _this = this;
        this.data.length = 0;
        (_a = this.data).push.apply(_a, data.map(function (item) { return ({ id: _this.getId(item), label: _this.getLabel(item) }); }));
    };
    TagInput.prototype.addTag = function (label) {
        return __awaiter(this, void 0, void 0, function () {
            var selectedTag, tag, newItem, tagTemplateData, tagHtml, selectedTagElement_1;
            var _this = this;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        selectedTag = this.currentSelection.find(function (t) { return t.label === label; });
                        if (!!selectedTag) return [3 /*break*/, 3];
                        tag = this.data.find(function (t) { return t.label === label; });
                        if (!!tag) return [3 /*break*/, 2];
                        // If this input shouldn't allow new tags, just return now
                        if (!this.allowNewTags) {
                            // TODO: make the UX a bit better (red text? callback function to handle error?)
                            return [2 /*return*/];
                        }
                        return [4 /*yield*/, this.newItemFactory(label)];
                    case 1:
                        newItem = _a.sent();
                        tag = {
                            id: this.getId(newItem),
                            label: this.getLabel(newItem)
                        };
                        // Add the tag to the internal data source, so that if it's deleted and re-added again
                        // then we get it from there next time and don't create another new one
                        this.data.push(tag);
                        _a.label = 2;
                    case 2:
                        this.currentSelection.push(tag);
                        tagTemplateData = {
                            id: tag.id,
                            label: tag.label,
                            globalCssClassPrefix: this.globalCssClassPrefix
                        };
                        tagHtml = Mustache.render(this.itemTemplate, tagTemplateData);
                        // Insert the new tag before the hidden input
                        this.tagInputHiddenInput.insertAdjacentHTML('beforebegin', tagHtml);
                        this.clearTagInput();
                        this.hideSuggestions();
                        this.updateHiddenInput(this.currentSelection);
                        this.onTagAdded(this, [tag], this.currentSelection);
                        this.onTagsChanged(this, [tag], [], this.currentSelection);
                        return [3 /*break*/, 4];
                    case 3:
                        selectedTagElement_1 = this.tagInputContainer
                            .querySelector(".".concat(this.globalCssClassPrefix, "-tag[data-label=\"").concat(selectedTag.label, "\"]"));
                        selectedTagElement_1.classList.add(this.warningClass);
                        this.clearTagInput();
                        setTimeout(function () {
                            selectedTagElement_1.classList.remove(_this.warningClass);
                        }, 1500);
                        _a.label = 4;
                    case 4: return [2 /*return*/];
                }
            });
        });
    };
    TagInput.prototype.getTextInputValue = function (allowUpperCase) {
        var value = this.tagInputTextInput.value.trim();
        return !allowUpperCase
            ? value.toLowerCase()
            : value;
    };
    TagInput.prototype.getHiddenInputValue = function () {
        return this.tagInputHiddenInput.value.trim();
    };
    TagInput.prototype.showTagInput = function () {
        if (this.placeholderText) {
            this.tagInputTextInput.setAttribute('placeholder', this.placeholderText);
        }
        // Remove the narrowing class, restoring input to its original width
        this.tagInputTextInput.classList.remove(this.narrowedInputClass);
    };
    TagInput.prototype.hideTagInput = function () {
        if (this.getTextInputValue(this.allowUpperCase) === EmptyString) {
            this.tagInputTextInput.setAttribute('placeholder', EmptyString);
            // When the tag text input loses focus, add a class which narrows it
            // to 1px wide. This is to avoid odd visual effects when the tags in
            // the control wrap onto multiple lines
            this.tagInputTextInput.classList.add(this.narrowedInputClass);
        }
    };
    TagInput.prototype.clearTagInput = function () {
        this.tagInputTextInput.value = null;
    };
    TagInput.prototype.updateHiddenInput = function (data) {
        this.tagInputHiddenInput.value = data.map(function (t) { return t.id; }).join(this.tagDataSeparator);
    };
    TagInput.prototype.showSuggestions = function () {
        this.tagInputSuggestionDropdown.classList.remove(this.suggestionDropdownHiddenClass);
    };
    TagInput.prototype.hideSuggestions = function () {
        this.tagInputSuggestionDropdown.classList.add(this.suggestionDropdownHiddenClass);
    };
    TagInput.prototype.removeTag = function (tagElement) {
        var label = tagElement.getAttribute('data-label');
        var idx = this.currentSelection.findIndex(function (i) { return i.label === label; });
        var removedTag = this.currentSelection[idx];
        this.currentSelection.splice(idx, 1);
        tagElement.remove();
        this.updateHiddenInput(this.currentSelection);
        this.onTagRemoved(this, [removedTag], this.currentSelection);
        this.onTagsChanged(this, [], [removedTag], this.currentSelection);
    };
    TagInput.prototype.searchTags = function (tags, query) {
        var _this = this;
        return tags
            .filter(function (t) { return !_this.currentSelection.includes(t); })
            .filter(function (t) { return t.label.toUpperCase().indexOf(query) > -1; })
            .map(function (t) { return [t.label.toUpperCase().indexOf(query), t]; })
            .sort(function (a, b) {
            // If the position of the substring match differs, sort by that
            if (a[0] !== b[0]) {
                return a[0] - b[0];
            }
            // Otherwise, sort alphabetically
            return a[1].label.localeCompare(b[1].label);
        })
            .map(function (f) { return f[1]; });
    };
    return TagInput;
}());
exports.TagInput = TagInput;
